#  Copyright (c) 2020. Sinri Edogawa


from abc import ABC
from typing import Iterable, List

import pymysql

from nehushtan.mysql import constant
from nehushtan.mysql.MySQLCondition import MySQLCondition
from nehushtan.mysql.MySQLQueryResult import MySQLQueryResult
from nehushtan.mysql.MySQLViewMixin import MySQLViewMixin


class MySQLTableMixin(MySQLViewMixin, ABC):

    def insert_one_row(
            self,
            row_dict: dict,
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        return self._write_one_row(row_dict, 'INSERT', commit_immediately, on_duplicate_key_update_rows, with_ignore)

    def replace_one_row(self, row_dict: dict, commit_immediately: bool = False):
        return self._write_one_row(row_dict, 'REPLACE', commit_immediately)

    @staticmethod
    def _build_on_duplicate_key_update_sql(on_duplicate_key_update_rows: dict):
        sql_parts = []
        for field, value in on_duplicate_key_update_rows.items():
            sql_parts.append(f'{field}={value}')
        sql_parts = ", ".join(sql_parts)
        return f'ON DUPLICATE KEY UPDATE {sql_parts}'

    def _write_one_row(
            self,
            row_dict: dict,
            write_type: str,
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        fields = []
        values = []
        for k, v in row_dict.items():
            fields.append('`' + k + '`')
            values.append(self.get_mysql_kit().quote(v))
        fields = ",".join(fields)
        values = ",".join(values)
        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'
        sql = f"{write_type}{ignore} INTO {self.get_table_expression()} ({fields}) VALUES ({values})"

        if type(on_duplicate_key_update_rows) is dict:
            sql_parts = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)
            sql = f'{sql} {sql_parts}'

        return self._modify_with_sql(sql, commit_immediately)

    def insert_many_rows_with_dicts(
            self,
            row_dicts: List[dict],
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        return self._write_many_rows_with_dicts(
            row_dict_array=row_dicts,
            write_type='INSERT',
            commit_immediately=commit_immediately,
            on_duplicate_key_update_rows=on_duplicate_key_update_rows,
            with_ignore=with_ignore
        )

    def replace_many_rows_with_dicts(self, row_dicts: List[dict], commit_immediately: bool = False):
        return self._write_many_rows_with_dicts(
            row_dict_array=row_dicts,
            write_type='REPLACE',
            commit_immediately=commit_immediately
        )

    def insert_many_rows_with_matrix(
            self,
            fields: Iterable[str],
            row_matrix: Iterable[Iterable],
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        return self._write_many_rows_with_matrix(
            fields=fields,
            row_matrix=row_matrix,
            write_type='INSERT',
            commit_immediately=commit_immediately,
            on_duplicate_key_update_rows=on_duplicate_key_update_rows,
            with_ignore=with_ignore
        )

    def replace_many_rows_with_matrix(
            self,
            fields: Iterable[str],
            row_matrix: Iterable[Iterable],
            commit_immediately: bool = False
    ):
        return self._write_many_rows_with_matrix(
            fields=fields,
            row_matrix=row_matrix,
            write_type='REPLACE',
            commit_immediately=commit_immediately
        )

    def _write_many_rows_with_dicts(
            self,
            row_dict_array: List[dict],
            write_type: str,
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        if len(row_dict_array) <= 0:
            return MySQLQueryResult.create_error_result('Rows Empty')

        sample_row = row_dict_array[0]

        fields = []
        for k, v in sample_row.items():
            fields.append(f'`{k}`')
        fields = ",".join(fields)

        row_sql = []
        for row in row_dict_array:
            values = []
            for k, v in row.items():
                values.append(self.get_mysql_kit().quote(v))
            values = ",".join(values)
            row_sql.append(f'({values})')
        if len(row_sql) <= 0:
            raise ValueError("You are trying to write zero rows!")
        row_sql = ",".join(row_sql)

        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'

        sql = f"{write_type}{ignore} INTO {self.get_table_expression()} ({fields}) VALUES {row_sql}"

        if type(on_duplicate_key_update_rows) is dict:
            sql_parts = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)
            sql = f'{sql} {sql_parts}'

        return self._modify_with_sql(sql, commit_immediately)

    def _write_many_rows_with_matrix(
            self,
            fields: Iterable[str],
            row_matrix: Iterable[Iterable],
            write_type: str,
            commit_immediately: bool = False,
            on_duplicate_key_update_rows: dict = None,
            with_ignore: bool = False
    ):
        fields_sql = ",".join([f"`{x}`" for x in fields])
        row_sql = []
        for row in row_matrix:
            values = []
            for v in row:
                values.append(self.get_mysql_kit().quote(v))
            values = ",".join(values)
            row_sql.append(f'({values})')
        if len(row_sql) <= 0:
            raise ValueError("You are trying to write zero rows!")
        row_sql = ",".join(row_sql)

        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'

        sql = f"{write_type}{ignore} INTO {self.get_table_expression()} ({fields_sql}) VALUES {row_sql}"

        if type(on_duplicate_key_update_rows) is dict:
            sql_parts = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)
            sql = f'{sql} {sql_parts}'

        return self._modify_with_sql(sql, commit_immediately)

    def write_rows_with_raw_selection_sql(
            self,
            write_type: str,
            fields: Iterable[str],
            selection_sql: str,
            with_ignore: bool = False,
            on_duplicate_key_update_rows: dict = None,
            commit_immediately: bool = False
    ):
        """
        < Experimental Method, Since 0.1.10>
        It is a low level method for
        [INSERT ... SELECT Statement](https://dev.mysql.com/doc/refman/8.0/en/insert-select.html)
        and [REPLACE Statement](https://dev.mysql.com/doc/refman/8.0/en/replace.html)
        Be Careful when use this.
        :param write_type: INSERT or REPLACE
        :param fields:
        :param selection_sql: 'SELECT ...' If select from one table, you may use MySQLTableSelection::generate_sql()
        :param with_ignore:
        :param on_duplicate_key_update_rows:
        :param commit_immediately:
        :return:
        """
        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'

        fields_sql = ",".join([f'`{x}`' for x in fields])

        sql = f"{write_type}{ignore} INTO {self.get_table_expression()} ({fields_sql}) {selection_sql}"

        if type(on_duplicate_key_update_rows) is dict:
            sql_parts = self._build_on_duplicate_key_update_sql(on_duplicate_key_update_rows)
            sql = f'{sql} {sql_parts}'

        return self._modify_with_sql(sql, commit_immediately)

    def _modify_with_sql(self, sql: str, commit_immediately: bool = False):
        result = MySQLQueryResult()
        cursor = None

        try:
            result.set_sql(sql)
            cursor = self.get_mysql_kit().get_raw_connection().cursor()
            cursor.execute(sql)
            result.set_status(constant.MYSQL_QUERY_STATUS_EXECUTED)
            result.set_last_inserted_id(cursor.lastrowid)
            result.set_affected_rows(cursor.rowcount)
            if commit_immediately:
                self.get_mysql_kit().get_raw_connection().commit()

        except pymysql.MySQLError as e:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"MySQL Error {e.__class__}: [{e.args[0]}] {e.args[1]}")

        except Exception as pe:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"Python Error {pe.__class__}: {pe}")

        finally:
            if cursor is not None:
                cursor.close()
            return result

    def update_rows(
            self,
            conditions: Iterable[MySQLCondition],
            modifications: dict,
            commit_immediately: bool = False,
            with_ignore: bool = False,
            sort_expression: str = '',
            limit: int = 0
    ):
        """
        [UPDATE Statement](https://dev.mysql.com/doc/refman/8.0/en/update.html)
        :param conditions:
        :param modifications:
        :param commit_immediately:
        :param with_ignore:
        :param sort_expression:
        :param limit:
        :return:
        """
        condition_sql = MySQLCondition.build_sql_component(conditions)

        modify_pairs = []
        for k, v in modifications.items():
            modify_pairs.append(f'`{k}`=' + self.get_mysql_kit().quote(v))
        modify_pairs = ",".join(modify_pairs)

        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'

        sql = f"UPDATE{ignore} {self.get_table_expression()} SET {modify_pairs} WHERE {condition_sql}"

        if len(sort_expression) > 0:
            sql = f"{sql} ORDER BY {sort_expression}"

        if limit > 0:
            sql = f"{sql} LIMIT {limit}"

        return self._modify_with_sql(sql, commit_immediately)

    def delete_rows(
            self,
            conditions: Iterable[MySQLCondition],
            commit_immediately: bool = False,
            with_ignore: bool = False,
            sort_expression: str = '',
            limit: int = 0
    ):
        """
        [DELETE Statement](https://dev.mysql.com/doc/refman/8.0/en/delete.html)
        :param conditions:
        :param commit_immediately:
        :param with_ignore:
        :param sort_expression:
        :param limit:
        :return:
        """
        condition_sql = MySQLCondition.build_sql_component(conditions)

        ignore = ''
        if with_ignore:
            ignore = ' IGNORE'

        sql = f"DELETE{ignore} FROM {self.get_table_expression()} WHERE {condition_sql}"

        if len(sort_expression) > 0:
            sql = f"{sql} ORDER BY {sort_expression}"

        if limit > 0:
            sql = f"{sql} LIMIT {limit}"

        return self._modify_with_sql(sql, commit_immediately)



#  Copyright (c) 2020. Sinri Edogawa
from typing import Iterable

import pymysql
from pymysql.cursors import SSCursor, SSDictCursor

from nehushtan.mysql import constant
from nehushtan.mysql.MySQLCondition import MySQLCondition
from nehushtan.mysql.MySQLQueryResult import MySQLQueryResult
from nehushtan.mysql.MySQLTableExistence import MySQLTableExistence


class MySQLTableSelection:
    # _model: MySQLTableExistence
    # _select_fields: list
    # _conditions: list
    # _group_by_fields: list
    # _sort_expression: str
    # _limit: int
    # _offset: int
    # _use_indices: list
    # _force_indices: list
    # _ignore_indices: list
    # _for_update: bool

    def __init__(self, model: MySQLTableExistence):
        self._model = model
        self._select_fields = []
        self._conditions = []
        self._group_by_fields = []
        self._sort_expression = ''
        self._limit = 0
        self._offset = 0
        self._use_indices = []
        self._force_indices = []
        self._ignore_indices = []
        self._for_update = False

    def get_limit(self) -> int:
        return self._limit

    def set_limit(self, limit: int):
        self._limit = limit
        return self

    def get_offset(self) -> int:
        return self._offset

    def set_offset(self, offset: int):
        self._offset = offset
        return self

    def get_sort_expression(self):
        return self._sort_expression

    def set_sort_expression(self, sort_expression: str):
        self._sort_expression = sort_expression
        return self

    def add_select_field(self, field_expression: str, alias: str = ''):
        s = field_expression
        if len(alias) > 0:
            s += f' as {alias}'
        self._select_fields.append(s)
        return self

    def add_select_field_name_list(self, field_name_array: Iterable[str]):
        """

        :param field_name_array: ["F1","F2", ...] or ("F1","F2", ...)
        :return:
        """
        for item in field_name_array:
            self._select_fields.append(item)
        return self

    def add_condition(self, condition: MySQLCondition):
        self._conditions.append(condition)
        return self

    def add_conditions(self, conditions: Iterable[MySQLCondition]):
        """

        :param conditions: array of MySQLCondition
        :return:
        """
        for condition in conditions:
            if isinstance(condition, MySQLCondition):
                self._conditions.append(condition)
            # else just ignore
        return self

    def add_simple_conditions(self, equal_dict: dict):
        """

        :param equal_dict: {FIELD:VALUE,FIELD:VALUE_ARRAY}
        :return:
        """
        for k, v in equal_dict.items():
            if isinstance(v, (tuple, list)):
                self.add_condition(MySQLCondition.make_in_array(k, v))
            else:
                self.add_condition(MySQLCondition.make_equal(k, v))
        return self

    def set_group_by_fields(self, group_by_fields: list):
        self._group_by_fields = group_by_fields
        return self

    def use_index(self, index: str):
        self._use_indices.append(index)
        return self

    def force_index(self, index: str):
        self._force_indices.append(index)
        return self

    def ignore_index(self, index: str):
        self._ignore_indices.append(index)
        return self

    def set_for_update(self, value: bool):
        self._for_update = value

    def generate_sql(self):
        table = self._model.get_table_expression()

        fields = "*"
        if len(self._select_fields) > 0:
            fields = ",".join(self._select_fields)

        condition_sql = MySQLCondition.build_sql_component(self._conditions)

        indices = ''
        if len(self._use_indices) > 0:
            indices += " USE INDEX (" + ",".join(self._use_indices) + ") "
        if len(self._force_indices) > 0:
            indices += " FORCE INDEX (" + ",".join(self._force_indices) + ") "
        if len(self._ignore_indices) > 0:
            indices += " IGNORE INDEX (" + ",".join(self._force_indices) + ") "

        sql = f"SELECT {fields} FROM {table} {indices} WHERE {condition_sql} "

        if len(self._group_by_fields) > 0:
            sql += "GROUP BY " + ",".join(self._group_by_fields) + " "

        if self._sort_expression.strip() != '':
            sql += "ORDER BY " + self._sort_expression + " "

        if self._limit > 0:
            sql += f"LIMIT {self._limit} "
            if self._offset > 0:
                sql += f"OFFSET {self._offset} "

        if self._for_update:
            sql += " FOR UPDATE "

        return sql

    def query_for_result_matrix(self, row_type: type):
        result = MySQLQueryResult()
        try:
            sql = self.generate_sql()
            result.set_sql(sql)
            if row_type is dict:
                matrix = self._model.get_mysql_kit().raw_query_for_all_dict_rows(sql)
            else:
                matrix = self._model.get_mysql_kit().raw_query_for_all_tuple_rows(sql)
            result.set_status(constant.MYSQL_QUERY_STATUS_QUERIED)
            result.append_result_rows(matrix)
        except pymysql.MySQLError as e:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"MySQL Error {e.__class__} [{e.args[0]}] {e.args[1]}")
        except Exception as pe:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"Python Error {pe.__class__}: {pe}")
        finally:
            return result

    def query_for_result_as_tuple_of_dict(self):
        return self.query_for_result_matrix(dict)

    def query_for_result_as_tuple_of_tuple(self):
        return self.query_for_result_matrix(tuple)

    def query_for_result_stream(self, row_type: type):
        result = MySQLQueryResult()
        try:
            sql = self.generate_sql()
            result.set_sql(sql)
            if row_type is dict:
                cursor = self._model.get_mysql_kit().get_raw_connection().cursor(SSDictCursor)
            else:
                cursor = self._model.get_mysql_kit().get_raw_connection().cursor(SSCursor)
            cursor.execute(sql)
            result.set_status(constant.MYSQL_QUERY_STATUS_STREAMING)
            result.set_stream(cursor)
            return result
        except pymysql.MySQLError as e:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"MySQL Error {e.__class__} [{e.args[0]}] {e.args[1]}")
        except Exception as pe:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"Python Error {pe.__class__}: {pe}")
        finally:
            return result

    def query_for_result_stream_as_dict(self):
        return self.query_for_result_stream(row_type=dict)

    def query_for_result_stream_as_tuple(self):
        return self.query_for_result_stream(row_type=tuple)



#  Copyright (c) 2020. Sinri Edogawa

from typing import Callable

import pymysql
from pymysql.connections import Connection
from pymysql.cursors import DictCursor

from nehushtan.mysql import constant
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig


class MySQLKit:
    """
    MySQL 连接基础工具类，封装了 PYMYSQL 库的实例。
    Shovel 项目相关的所有 MySQL 连接应当使用此方法进行。
    Since 0.1.13 Added Auto Reconnection (PING)
    """

    # _mysql_config: MySQLKitConfig
    # _auto_commit_default: bool
    # _connection: Connection or None

    def __init__(self, mysql_config: MySQLKitConfig = None):
        """
        Since 0.1.4 入参类型变化
        获取一个MySQLKitConfig类的实例，并根据此类设置的配置立即建立一个新的数据库连接。
        PYMYSQL 默认不自动递交。在实际场景中，建议设置成默认自动递交并仅在需要时手动开启事务。
        :param mysql_config: 获取一个MySQLKitConfig类的实例，可用字典构建，样本参见配置文件标准示例中 `mysql_database`.`sample` 项
        """
        if mysql_config is None:
            mysql_config = MySQLKitConfig()
        self._mysql_config = mysql_config
        self._auto_commit_default = self._mysql_config.get_auto_commit()
        self._connection = None
        if len(self._mysql_config.get_host()) > 0:
            self.connect()

    @staticmethod
    def make_instance_from_pymysql_connection(connection: Connection):
        """
        根据一个已经存在的 PYMYSQL 连接，构建一个本类实例。
        这样建立的实例里的配置字典为空，而自动提交参数根据入参连接的实际情况获取。
        :param connection:
        :return: 本类实例
        """
        kit = MySQLKit()
        kit._connection = connection
        kit._auto_commit_default = connection.get_autocommit()
        return kit

    def get_raw_connection(self) -> Connection:
        """
        :return: 封装的 PYMYSQL 原生 CONNECTION 实例
        """
        if self._connection is None:
            raise RuntimeError("Connection Property of MySQLKit Instance is not set.")
        self._connection.ping(reconnect=True)
        return self._connection

    def connect(self):
        """
        It may raise Exception if connection failed, as I viewed the document
        :return: MySQLKit the instance itself
        """

        self.disconnect()

        self._auto_commit_default = self._mysql_config.get_auto_commit()
        self._connection = pymysql.connect(
            host=self._mysql_config.get_host(),  # '172.16.1.52',
            port=self._mysql_config.get_port(),  # 3306,
            user=self._mysql_config.get_user(),
            password=self._mysql_config.get_password(),
            db=self._mysql_config.get_db(),
            charset=self._mysql_config.get_charset(),
            autocommit=self._auto_commit_default,
        )
        return self

    def disconnect(self):
        if self._connection is not None:
            self._connection.close()
        self._connection = None
        return self

    def raw_get_computed_sql(self, sql: str, args=None) -> str:
        """
        使用 PYMYSQL 原生的方法获取构建的 SQL 语句用于预览
        :param sql: SQL 模板
        :param args: 参数
        :return: 生成的 SQL 语句
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()
        sql = cursor.mogrify(sql, args)
        cursor.close()
        return sql

    def raw_query_for_all_tuple_rows(self, sql: str, args=None) -> tuple:
        """
        查询一个SQL并获取包含所有结果行的元组，其中的每行也封装成一个元组
        :param sql:
        :param args:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()
        cursor.execute(sql, args)
        rows_tuple = cursor.fetchall()
        cursor.close()
        return rows_tuple

    def raw_query_for_all_dict_rows(self, sql: str, args=None) -> tuple:
        """
        查询一个SQL并获取包含所有结果行的元组，其中的每行封装成一个字典
        :param sql:
        :param args:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor(DictCursor)
        cursor.execute(sql, args)
        rows_tuple = cursor.fetchall()
        cursor.close()
        return rows_tuple

    def raw_query_to_modify_one(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行一个非读SQL并返回影响行数
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()
        afx = cursor.execute(sql, args)
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return afx

    def raw_query_to_insert_one(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行一条插入语句并获取最后更新的行的ID
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()
        cursor.execute(sql, args)
        last_row_id = cursor.lastrowid
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return last_row_id

    def raw_query_to_modify_many(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行原生的SQL批量变更并获取影响行数
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()
        afx = cursor.executemany(sql, args)
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return afx

    def raw_query_to_insert_many(self, sql: str, args=None, commit_immediately: bool = None):
        """
        执行原生的SQL批量插入并获取最后更新行的ID
        :param sql:
        :param args:
        :param commit_immediately:
        :return:
        """
        connection = self.get_raw_connection()
        cursor = connection.cursor()
        cursor.executemany(sql, args)
        last_row_id = cursor.lastrowid
        if commit_immediately is None:
            commit_immediately = self._auto_commit_default
        if commit_immediately:
            connection.commit()
        cursor.close()
        return last_row_id

    def raw_execute_transaction(self, transaction_callable: Callable[["MySQLKit"], any]):
        """
        一个封装的方法，用于快速包装事务。
        :param transaction_callable: 事务处理具体代码所在的闭包，可以使用lambda表达式或者def方法，参数为本类实例，并返回结果
        :return: 返回闭包的返回值
        """
        self.get_raw_connection().begin()
        try:
            result = transaction_callable(self)
            print('to commit', result)
            self._connection.commit()
            return result
        except Exception as e:
            print('to rollback')
            self._connection.rollback()
            raise e

    def quote(self, value):
        """
        将任意入参进行转义以防范注入。
        如果连接关闭，则使用脱机转义。
        数字、空值、布尔值不加引号，其他都作为字符串加外部引号可用于字符串拼接。
        如果入参为元组或者列表，也以同样形式返回转义所有元素的新实例。
        :param value:
        :return:
        """
        if self._connection is None:
            return self.quote_offline(value)
        self._connection.ping(True)

        if type(value) == list:
            x = []
            for item in value:
                x.append(self.quote(item))
            return x
        elif type(value) == tuple:
            x = []
            for item in value:
                x.append(self.quote(item))
            return tuple(x)
        elif type(value) == int or type(value) == float:
            # Fixed in 0.1.8: if number is returned, the str plus would fail
            return f"{value}"
        elif value is None:
            return constant.MYSQL_CONDITION_CONST_NULL
        elif value is True:
            return constant.MYSQL_CONDITION_CONST_TRUE
        elif value is False:
            return constant.MYSQL_CONDITION_CONST_FALSE
        elif type(value) == str:
            escaped = self._connection.escape_string(value)
            return f"'{escaped}'"
        else:
            return self.quote(value.__str__())

    @staticmethod
    def quote_offline(value):
        """
        脱机转义
        :param value:
        :return:
        """
        if type(value) == list:
            x = []
            for item in value:
                x.append(MySQLKit.quote_offline(item))
            return x
        elif type(value) == tuple:
            x = []
            for item in value:
                x.append(MySQLKit.quote_offline(item))
            return tuple(x)
        elif type(value) == int or type(value) == float:
            return f'{value}'
        elif type(value) == str:
            return MySQLKit.quote_string_offline(value)
        elif value is None:
            return constant.MYSQL_CONDITION_CONST_NULL
        elif value is True:
            return constant.MYSQL_CONDITION_CONST_TRUE
        elif value is False:
            return constant.MYSQL_CONDITION_CONST_FALSE
        else:
            return MySQLKit.quote_string_offline(value.__str__())

    @staticmethod
    def quote_string_offline(value: str) -> str:
        """
        脱机转义字符串。
        :param value:
        :return: 包含了外层引号
        """
        d = {
            '\\': '\\\\',
            '\0': '\\0',
            '\n': '\\n',
            '\r': '\\r',
            "'": "\\'",
            '"': '\\"',
            '\x1a': '\\Z',
        }
        s = value
        for k, v in d.items():
            s = s.replace(k, v)

        return f"'{s}'"



import re


class NehushtanArgumentParser:

    def __init__(self, sys_arg_list: list):
        """

        :param sys_arg_list: Commonly `sys.argv[1:]`
        """
        self._sys_arg_list = sys_arg_list
        self._argument_dict = {}
        self._options_short = {}
        self._options_long = {}
        self._options_prefix = {}
        self._result_dict = {}
        self._usage_dict = {}

    def add_option(self, key: str, desc: str = '', short: str = '', long: str = '', prefix: str = ''):
        """

        :param desc:
        :param key: Any Name
        :param short: a -> `-a`
        :param long: alpha -> `--alpha`
        :param prefix: alef -> `--alef-`
        :return:
        """
        if short == '' and long == '' and prefix == '':
            return self

        usage = {
            'key': key,
            'desc': desc,
        }
        if short != '':
            self._options_short[f'-{short}'] = key
            usage['short'] = f'-{short} [VALUE]'
        if long != '':
            self._options_long[f'--{long}'] = key
            usage['long'] = f'--{long} [VALUE]'
        if prefix != '':
            self._options_prefix[f'--{prefix}-'] = key
            usage['prefix'] = f'--{prefix}-[KEY] [VALUE]'

        self._usage_dict[key] = usage
        return self

    def parse(self):
        current_key = None
        self._argument_dict = {}
        for item in self._sys_arg_list:
            if current_key is None:
                current_key = item
            else:
                self._argument_dict[current_key] = item
                current_key = None

        self._result_dict = {}
        for (key, value) in self._argument_dict.items():
            if self._options_short.get(key) is not None:
                self._result_dict[self._options_short[key]] = value
            elif self._options_long.get(key) is not None:
                self._result_dict[self._options_long[key]] = value
            else:
                for prefix, mapping_key in self._options_prefix.items():
                    matches = re.match(f'^{prefix}([A-Za-z0-9._-]+)$', key)
                    if matches is not None:
                        if self._result_dict.get(mapping_key) is None:
                            self._result_dict[mapping_key] = {}
                        self._result_dict[mapping_key][matches[1]] = value
                        break
        return self

    def get_raw_option_dict(self):
        return self._argument_dict

    def get_parsed_option_dict(self):
        return self._result_dict

    def get_usage_text(self):
        usage_text = ""
        for key, value in self._usage_dict.items():
            options = []
            if len(value.get('short', '')) > 0:
                options.append(value.get('short'))
            if len(value.get('long', '')) > 0:
                options.append(value.get('long'))
            if len(value.get('prefix', '')) > 0:
                options.append(value.get('prefix'))

            usage_text += f"{value['key']}" + "\t\t"
            usage_text += ", ".join(options)
            if len(value.get('desc', '')) > 0:
                usage_text += f"\n{value['desc']}"
            usage_text += "\n"

        return usage_text



#  Copyright (c) 2020. Sinri Edogawa

MYSQL_QUERY_STATUS_INIT = "INIT"
MYSQL_QUERY_STATUS_QUERIED = "QUERIED"
MYSQL_QUERY_STATUS_EXECUTED = "EXECUTED"
MYSQL_QUERY_STATUS_STREAMING = "STREAMING"
MYSQL_QUERY_STATUS_STREAMED = "STREAMED"
MYSQL_QUERY_STATUS_ERROR = "ERROR"



#  Copyright (c) 2020. Sinri Edogawa

MYSQL_CONDITION_OP_EQ = "="
MYSQL_CONDITION_OP_GT = ">"
MYSQL_CONDITION_OP_EGT = ">="
MYSQL_CONDITION_OP_LT = "<"
MYSQL_CONDITION_OP_ELT = "<="
MYSQL_CONDITION_OP_NEQ = "<>"
MYSQL_CONDITION_OP_NULL_SAFE_EQUAL = "<=>"
MYSQL_CONDITION_OP_IS = "IS"
MYSQL_CONDITION_OP_IS_NOT = "IS NOT"
MYSQL_CONDITION_OP_IN = "IN"
MYSQL_CONDITION_OP_NOT_IN = "NOT IN"
MYSQL_CONDITION_OP_LIKE = "LIKE"
MYSQL_CONDITION_OP_NOT_LIKE = "NOT LIKE"
MYSQL_CONDITION_OP_BETWEEN = "BETWEEN"
MYSQL_CONDITION_OP_NOT_BETWEEN = "NOT BETWEEN"

MYSQL_CONDITION_MACRO_IS_NULL_OR_EMPTY_STRING = "MACRO_IS_NULL_OR_EMPTY_STRING"
MYSQL_CONDITION_MACRO_IS_NOT_NULL_NOR_EMPTY_STRING = "MACRO_IS_NOT_NULL_NOR_EMPTY_STRING"

MYSQL_CONDITION_MACRO_RAW_EXPRESSION = "MACRO_RAW_EXPRESSION"

MYSQL_CONDITION_OP_EXISTS = "EXISTS"
MYSQL_CONDITION_OP_NOT_EXISTS = "NOT EXISTS"

MYSQL_CONDITION_OP_PARENTHESES_AND = "AND"
MYSQL_CONDITION_OP_PARENTHESES_OR = "OR"

MYSQL_CONDITION_CONST_TRUE = "TRUE"
MYSQL_CONDITION_CONST_FALSE = "FALSE"
MYSQL_CONDITION_CONST_NULL = "NULL"

MYSQL_CONDITION_LIKE_LEFT_WILDCARD = "LIKE_LEFT_WILDCARD"
MYSQL_CONDITION_LIKE_RIGHT_WILDCARD = "LIKE_RIGHT_WILDCARD"
MYSQL_CONDITION_LIKE_BOTH_WILDCARD = "LIKE_BOTH_WILDCARD"



class MySQLKitConfig:
    """
    A standard config class for MySQLKit
    Since 0.1.4
    """

    # _host: str
    # _port: int
    # _user: str
    # _password: str
    # _db: str
    # _charset: str
    # _auto_commit: bool

    def __init__(self, config_dict: dict = None):
        if config_dict is None:
            config_dict = {}

        self._host = config_dict.get('host', '')
        self._port = config_dict.get('port', 3306)
        self._user = config_dict.get('user', 'anonymous')
        self._password = config_dict.get('password', '')
        self._db = config_dict.get('db', '')
        self._charset = config_dict.get('charset', 'utf8')
        self._auto_commit = config_dict.get('auto_commit', True)

    def get_host(self):
        return self._host

    def set_host(self, host: str):
        self._host = host
        return self

    def get_port(self):
        return self._port

    def set_port(self, port: int):
        self._port = port
        return self

    def get_user(self):
        return self._user

    def set_user(self, user: str):
        self._user = user
        return self

    def get_password(self):
        return self._password

    def set_password(self, password: str):
        self._password = password
        return self

    def get_db(self):
        return self._db

    def set_db(self, db: str):
        self._db = db
        return self

    def get_charset(self):
        return self._charset

    def set_charset(self, charset: str):
        self._charset = charset
        return self

    def get_auto_commit(self):
        return self._auto_commit

    def set_auto_commit(self, auto_commit: bool):
        self._auto_commit = auto_commit
        return self



#  Copyright (c) 2020. Sinri Edogawa
import warnings
from typing import Iterable

from pymysql.cursors import Cursor

from nehushtan.mysql import constant


class MySQLQueryResult:
    # _sql: str
    # _status: str
    # _error: str

    # The beneath properties would be initialized for each instance when needed
    # _last_inserted_id: int
    # _affected_rows: int
    # _result_rows: list
    # _result_stream: Cursor  # do not know how to implement it in python now

    def __init__(self):
        self._sql = ''
        self._status = constant.MYSQL_QUERY_STATUS_INIT
        self._error = 'Not executed yet!'

        self._last_inserted_id = -1
        self._affected_rows = -1
        self._result_rows = []
        self._result_stream = None

    def is_queried(self):
        return self._status == constant.MYSQL_QUERY_STATUS_QUERIED

    def is_executed(self):
        return self._status == constant.MYSQL_QUERY_STATUS_EXECUTED

    def is_streamed(self):
        return self._status == constant.MYSQL_QUERY_STATUS_STREAMED

    def get_last_inserted_id(self):
        return self._last_inserted_id

    def set_last_inserted_id(self, last_inserted_id: int):
        self._last_inserted_id = last_inserted_id
        return self

    def get_affected_rows(self):
        return self._affected_rows

    def set_affected_rows(self, affected_rows: int):
        self._affected_rows = affected_rows
        return self

    def get_sql(self):
        return self._sql

    def set_sql(self, sql: str):
        self._sql = sql
        return self

    def get_status(self):
        return self._status

    def set_status(self, status):
        self._status = status
        if (
                constant.MYSQL_QUERY_STATUS_QUERIED,
                constant.MYSQL_QUERY_STATUS_STREAMING,
                constant.MYSQL_QUERY_STATUS_STREAMED,
                constant.MYSQL_QUERY_STATUS_EXECUTED,
        ).__contains__(self._status):
            self._error = 'NO ERROR'
        return self

    def get_error(self):
        return self._error

    def set_error(self, error: str):
        self._error = error
        return self

    def append_result_rows(self, rows: Iterable):
        """

        :param rows: Array (list ot tuple) of result rows, each row would be a dict or tuple
        :return:
        """
        if type(rows) is tuple:
            rows = list(rows)
        self._result_rows += rows
        return self

    def set_stream(self, cursor: Cursor):
        self._result_stream = cursor
        return self

    def get_stream(self) -> Cursor:
        if self._result_stream is None:
            raise Exception("No available stream!")
        return self._result_stream

    def read_next_row(self):
        if self._status != constant.MYSQL_QUERY_STATUS_STREAMING:
            raise Exception("Now no stream to read!")
        row = self.get_stream().fetchone()
        if row is None:
            self._status = constant.MYSQL_QUERY_STATUS_STREAMED
            self._result_stream.close()
            del self._result_stream
            return None
        return row

    def get_result(self):
        warnings.warn('Deprecated since 0.1.12. Use `get_fetched_rows_as_tuple` instead.', DeprecationWarning)
        if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
            raise Exception('Cannot fetch query result as status is not QUERIED.')
        return self._result_rows

    def get_fetched_rows_as_tuple(self):
        if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
            raise Exception('Cannot fetch query result as status is not QUERIED.')
        return tuple(self._result_rows)

    @staticmethod
    def create_error_result(error_message: str):
        return MySQLQueryResult().set_status(constant.MYSQL_QUERY_STATUS_ERROR).set_error(error_message)



#  Copyright (c) 2020. Sinri Edogawa

import json
import logging
import logging.handlers
import sys
from typing import Iterable


class NehushtanLogger:
    """
    A Logger Class For Shovel
    """

    def __init__(
            self,
            logger_name: str,
            handlers: Iterable[logging.Handler] = None,
            universal_log_level=logging.DEBUG,
            with_process_info=False,
            with_thread_info=False
    ):
        """
        Reusable Logger, determined by name
        If parameter `handlers` provided and not None, all previous handlers would be cleared and newly given ones set.
        """
        self.with_process_info = with_process_info
        self.with_thread_info = with_thread_info
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(universal_log_level)
        if handlers is not None:
            self.logger.handlers=[]
            for handler in handlers:
                # https://docs.python.org/3/library/logging.html#formatter-objects

                format_string = "%(asctime)s <%(name)s> [%(levelname)s]"
                if self.with_process_info:
                    format_string += " <%(process)d:%(processName)s>"
                if self.with_thread_info:
                    format_string += " <%(thread)d:%(threadName)s>"
                format_string += " %(message)s | %(json_string)s"

                handler.setFormatter(
                    logging.Formatter(format_string)
                )
                self.logger.addHandler(handler)

    def debug(self, message: str, extra=None):
        self.logger.debug(
            msg=message,
            extra=self.ensure_extra_as_dict(extra)
        )

    def info(self, message: str, extra=None):
        self.logger.info(
            msg=message,
            extra=self.ensure_extra_as_dict(extra)
        )

    def warning(self, message: str, extra=None):
        self.logger.warning(
            msg=message,
            extra=self.ensure_extra_as_dict(extra)
        )

    def error(self, message: str, extra=None):
        self.logger.error(
            msg=message,
            extra=self.ensure_extra_as_dict(extra)
        )

    def exception(self, message: str, exception: BaseException):
        """
        Since version 1.2 彭启航之野望
        :param message:
        :param exception:
        :return:
        """
        self.logger.exception(
            msg=message,
            extra=self.ensure_extra_as_dict(exception.__str__()),
            stack_info=True,
        )

    def critical(self, message: str, extra=None):
        self.logger.critical(
            msg=message,
            extra=self.ensure_extra_as_dict(extra)
        )

    @staticmethod
    def make_silent_handler():
        s_handler = logging.NullHandler()
        return s_handler

    @staticmethod
    def make_stdout_handler(logger_level=logging.DEBUG):
        s_handler = logging.StreamHandler(stream=sys.stdout)
        s_handler.setLevel(logger_level)
        return s_handler

    @staticmethod
    def make_stderr_handler(logger_level=logging.WARN):
        s_handler = logging.StreamHandler(stream=sys.stderr)
        s_handler.setLevel(logger_level)
        return s_handler

    @staticmethod
    def make_timed_rotating_file_handler(file_name='shovel.log', logger_level=logging.DEBUG, backup_count=7):
        rf_handler = logging.handlers.TimedRotatingFileHandler(
            filename=file_name,
            when='midnight',
            interval=1,
            backupCount=backup_count
        )
        rf_handler.setLevel(logger_level)
        return rf_handler

    @staticmethod
    def make_fixed_file_handler(file_name='shovel.log', logger_level=logging.DEBUG):
        f_handler = logging.FileHandler(file_name)
        f_handler.setLevel(logger_level)
        return f_handler

    def ensure_extra_as_dict(self, extra):
        return {
            "json_string": json.dumps(extra, default=lambda inner_x: inner_x.__str__()),
        }

    @staticmethod
    def get_silent_logger(logger_name: str):
        return NehushtanLogger(
            logger_name=logger_name,
            handlers=[NehushtanLogger.make_silent_handler()]
        )



class CommonHelper:
    def __init__(self):
        pass

    @staticmethod
    def read_target(target, keychain: tuple, default: any = None):
        if type(target) is dict:
            return CommonHelper.read_dictionary(target, keychain, default)
        elif type(target) is list or type(target) is tuple:
            return CommonHelper.read_array(target, keychain, default)
        else:
            return default

    @staticmethod
    def read_dictionary(dictionary: dict, keychain: tuple, default: any = None):
        if len(keychain) <= 0:
            return default
        elif len(keychain) == 1:
            return dictionary.get(keychain[0], default)
        else:
            current_key = keychain[0]
            if dictionary.keys().__contains__(current_key):
                sub_dictionary = dictionary.get(current_key, None)
                if type(sub_dictionary) is dict:
                    return CommonHelper.read_dictionary(sub_dictionary, keychain[1:], default)
                elif type(sub_dictionary) is tuple or type(sub_dictionary) is list:
                    return CommonHelper.read_array(sub_dictionary, keychain[1:], default)
                else:
                    return default
            else:
                return default

    @staticmethod
    def read_array(array: tuple, keychain: tuple, default: any = None):
        if len(keychain) <= 0:
            return default
        elif len(keychain) == 1:
            current_key = int(keychain[0])
            if len(array) > current_key:
                return array[int(keychain[0])]
            else:
                return default
        else:
            current_key = int(keychain[0])
            if len(array) > current_key:
                sub_array = array[current_key]
                if type(sub_array) is tuple or type(sub_array) is list:
                    return CommonHelper.read_array(sub_array, keychain[1:], default)
                elif type(sub_array) is dict:
                    return CommonHelper.read_dictionary(sub_array, keychain[1:], default)
                else:
                    return default
            else:
                return default

    @staticmethod
    def write_dictionary(target_dict: dict, keychain: tuple, value: any):
        if len(keychain) <= 0:
            # not modify anything if keychain is empty
            return target_dict
        if len(keychain) == 1:
            target_dict[keychain[0]] = value
            return target_dict

        current_key = keychain[0]
        if target_dict.get(current_key) is None:
            target_dict[current_key] = {}
        return CommonHelper.write_dictionary(target_dict[current_key], keychain[1:], value)
